/*
 
  copyright (c) 2004, Michal Salaban <emes@pld-linux.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License, version 2 as
  published by the Free Software Foundation (see file COPYING for details).

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
*/

/* $Id: mem.c,v 1.4 2004/01/29 22:58:36 emes Exp $ */

#include "parseopt/parseopt.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <strings.h>

#define         BUFSIZE     512

#define OPT_FREE    1
#define OPT_USED    2
#define OPT_SWFREE  3
#define OPT_SWUSED  4
#define OPT_STEP    5
#define OPT_RUN     6
#define OPT_ONCE    7

#define OPT_NONE    0

optrepr     repr[] = {
    {   OPT_FREE,   O_SIZE,     "memfree"   },
    {   OPT_USED,   O_SIZE,     "memused"   },
    {   OPT_SWFREE, O_SIZE,     "swapfree"  },
    {   OPT_SWUSED, O_SIZE,     "swapused"  },
    {   OPT_STEP,   O_SIZE,     "step"      },
    {   OPT_RUN,    O_STRING,   "run"       },
    {   OPT_ONCE,   O_BOOL,     "once"      },
    {   OPT_NONE,   0,          NULL        }
};

optval      val[] = {
    {   OPT_FREE,   NULL    },
    {   OPT_USED,   NULL    },
    {   OPT_SWFREE, NULL    },
    {   OPT_SWUSED, NULL    },
    {   OPT_STEP,   NULL    },
    {   OPT_RUN,    NULL    },
    {   OPT_ONCE,   NULL    },
    {   OPT_NONE,   NULL    }
};

const char      *memfile = "/proc/meminfo";
const char      *usage  =
                "Usage: sysfence-mem OPTION=VALUE ... \n"
                "\n"
                "Where OPTION can be:\n"
                "\n"
                "   memfree     free memory bottom limit (bytes)\n"
                "   memused     used memory upper limit (bytes)\n"
                "   swapfree    swap free space bottom limit (bytes)\n"
                "   swapused    swap space in use upper limit (bytes)\n"
                "   step        how often probe for stats (seconds, default = 10)\n"
                "   run         command to run when memory usage exceeds thresholds\n"
                "   once        on  = invoke command only once after threshold "
                "has been exceeded\n"
                "               off = run command every step seconds if "
                "threshold has been exceeded\n"
                "               (use: on/off, 1/0, yes/no values) default is on\n"
                "\n"
                "The 'run' option is MANDATORY.\n"
                "At least one of mem*/swap* options must be specified. You may use k,M,G there\n"
                "(e.g. memfree=1k has the same meaning as memfree=1024).\n"
                "\n"
                "Report bugs to <emes@pld-linux.org>\n"
                "\n";

                    /* limits and current stats */
unsigned long int   l_mfree, l_mused, l_sfree, l_sused,
                    s_mfree, s_mused, s_sfree, s_sused;

char            *command = NULL;
char            fbuf[BUFSIZE];
int             fha;
long int        step = 10;
int             once = 1;

/* fast text parsing helpers */
char * to_newline (char *b)
{
    do {
        b++;
    } while ((*b != '\n') && 
             (*b != '\r') && 
             (*b != '\0'));
    /* skip to new line */
    if (*b != '\0') b++;
    return b;
}

char * to_notspace (char *b)
{
    do {
        b++;
    } while ((*b == ' ') || 
             (*b == '\t') || 
             (*b == '\n') || 
             (*b == '\r'));
    return b;
}

char * to_space (char *b)
{
    do {
        b++;
    } while ((*b != ' ') && 
             (*b != '\t') && 
             (*b != '\n') && 
             (*b != '\r') &&
             (*b != '\0'));
    return b;
}

void main_loop ()
{
    int                 tmp;
    int                 i;
    int                 done = 0, hit = 0;
    char                *ptr;
    unsigned long int   vmtot, vmused, vmbuf, vmcache,
                        vsused, vsfree;

    while (1) {
        hit = 0;
        
        lseek (fha, 0, SEEK_SET);
        read (fha, (char *) &fbuf, BUFSIZE);

        /* scanning /proc/meminfo for needed values */

        /* 1. memory total is in second line after 'Mem:\ *' */
        ptr = to_notspace (to_space (to_newline ((char *)&fbuf)));
        vmtot = atol (ptr);
        /* 2. used mem is next value */
        ptr = to_notspace (to_space (ptr));
        vmused = atol (ptr);
        /* 3. then there are: 'free' field, 'shared' field,
         *    and what we're looking for - buffers
         */
        ptr = to_notspace (to_space (
                    to_notspace (to_space (
                            to_notspace (to_space (ptr))
                            ))
                    ));
        vmbuf = atol (ptr);
        /* 4. cached is next field */
        ptr = to_notspace (to_space (ptr));
        vmcache = atol (ptr);

        /* 5. next line, swap values. first is 'Swap:', then total.
         *    next one is swap-used */
        ptr = to_notspace (to_space (
                    to_notspace (to_space (
                            to_newline (ptr)
                            ))
                    ));
        s_sused = atol (ptr);
        /* 6. swap-free */
        ptr = to_notspace (to_space (ptr));
        s_sfree = atol (ptr);

        /* calculate real values, i.e. substract buffers and cache size */
        s_mused = vmused - (vmbuf + vmcache);
        s_mfree = vmtot - s_mused;

        if (((l_mfree != 0) && (s_mfree < l_mfree)) ||
            ((l_mused != 0) && (s_mused > l_mused)) ||
            ((l_sfree != 0) && (s_sfree < l_sfree)) ||
            ((l_sused != 0) && (s_sused > l_sused))) {
                hit = 1;
                if ((done == 0) || (once == 0)) system (command);
        }
#ifdef DEBUG
        printf ("%ld: hit=%d done=%d, step=%d once=%d, "
                "now_memfree=%d now_memused=%d, "
                "now_swapfree=%d now_swapused=%d, "
                "now_memfree=%d now_memused=%d, "
                "now_swapfree=%d now_swapused=%d",
                time (0),
                hit, done,
                step, once,
                s_mfree, s_mused, s_sfree, s_sused,
                l_mfree, l_mused, l_sfree, l_sused
                );
        if ((done == 0) || (once == 0)) printf (", executed: %s", command);
        printf ("\n");
#endif
        
        done = hit;
        
        sleep (step);
    }
}

int main (int argc, char *argv[])
{
    int     i;
    int     cbs = 0, ccs = 0, slen;
    void    *o_mfree, *o_mused, *o_sfree, *o_sused, *stepopt, *onceopt;
    
    if (! fetch_opt_array (&val, &repr, OPT_NONE, argc - 1, &argv[1])) {
        fprintf (stderr, usage);
        exit (1);
    }

    o_mfree = get_val (&val, OPT_NONE, OPT_FREE);
    o_mused = get_val (&val, OPT_NONE, OPT_USED);
    o_sfree = get_val (&val, OPT_NONE, OPT_SWFREE);
    o_sused = get_val (&val, OPT_NONE, OPT_SWUSED);
    stepopt = get_val (&val, OPT_NONE, OPT_STEP);
    command = get_val (&val, OPT_NONE, OPT_RUN);
    onceopt = get_val (&val, OPT_NONE, OPT_ONCE);

    /* command and at least one of thresholds must be given */
    if (((o_mfree == NULL) &&
         (o_mused == NULL) &&
         (o_sfree == NULL) &&
         (o_sused == NULL)) ||
         (command == NULL)) {
        
        fprintf (stderr, usage);
        exit (1);
    }
    
    if (o_mfree != NULL) l_mfree = *((unsigned long int *) o_mfree);
    else l_mfree = 0;
    if (o_mused != NULL) l_mused = *((unsigned long int *) o_mused);
    else l_mused = 0;
    if (o_sfree != NULL) l_sfree = *((unsigned long int *) o_sfree);
    else l_sfree = 0;
    if (o_sused != NULL) l_sused = *((unsigned long int *) o_sused);
    else l_sused = 0;

    if (stepopt != NULL) step = *((long int *) stepopt);
    if (onceopt != NULL) once = *((int *) onceopt);
    
    /* open memfile (usually /proc/meminfo) */
    fha = open (memfile, O_RDONLY);
    if (fha < 0) {
        perror (memfile);
        exit (2);
    }
    
#ifndef DEBUG
    if (daemon (0, 0) == 0) main_loop ();
    else {
        perror ("daemon() failed");
        exit (3);
    }
#endif
#ifdef DEBUG
    main_loop ();
#endif

    return 0;
}

/* $Id: mem.c,v 1.4 2004/01/29 22:58:36 emes Exp $ */
